!----------------------------------------------------------
!   rsolution de l'quation de convection linaire 2d 
!      (sous forme de loi de conservation d/dt f + div(v f)=0
!   -  champ de vitesse variable donn
!   - sur maillage non structur de triangles
!   - schma volumes finis cell-centered, dcentr d'ordre 1, euler explicite en temps
!----------------------------------------------------------

program conv2d_non_structure_cc

  !----------- sous programmes ------------------------------------ 
  use mod_precision   !- (double ou simple prcision)
  use mod_vitesse     !- (champ de vitesse)
  use mod_ci          !- (donne initiale)
  use mod_aplusmoins  !- (calcul de a+ et a-)
  use mod_sorties     !- rsultats de calculs
  !----------------------------------------------------------------

  implicit none

  !------------ variables -----------------------------------------
  integer :: i,j,nb_noeuds,nb_trig,n,k,arete,tri1,tri2,iA,iB,nb_cotes,n1,n2,n3
  integer :: nmax,pas_affichage,nfich,sf
  character(len=100) :: fich,numero
  real(pr) :: AB,BC,AC,h,fb0,fb1,cfl,invdt,dt,R,fj,fi,vn,vnplus,vnmoins,temps
  real(pr) :: vmax,tmax,svnp
  real(pr), dimension(1:2) :: A,B,C,v,nij
  integer, dimension(:), allocatable :: reft,refs
  integer, dimension(:,:), allocatable :: s,e,ar,trig,flag
  real(pr), dimension(:), allocatable :: aire,diam,xc,yc,f,fnpun,flux,vnp,vnm
  real(pr), dimension(:,:), allocatable :: p,normale
  !----------------------------------------------------------------


  !------------- programme ----------------------------------------


  !--- lecture du maillage (format am_fmt donn par le mailleur EMC2)
  open(1,file='maillage.am_fmt',form='formatted',status='old')
  read (1,*) nb_noeuds,nb_trig    ! nombre de noeuds (sommets) et nombre de triangles
  allocate(p(nb_noeuds,2))        ! coordonnes de chaque noeud
  allocate(s(nb_trig,3))          ! numros des sommets de chaque triangle
  allocate(reft(nb_trig))         ! rfrence de chaque triangle (pour les CL) : non utilis ici
  allocate(refs(nb_noeuds))       ! rfrence de chaque noeud (pour les CL)
  read (1,*) ((s(j,i),i=1,3),j=1,nb_trig)
  read (1,*) ((p(j,i),i=1,2),j=1,nb_noeuds)
  read (1,*) ( reft(i),i=1,nb_trig)
  read (1,*) ( refs(i),i=1,nb_noeuds)
  close(1)

  !--- calcul de la connectivit
  
  !1)-calcul du nombre d'artes nb_cotes
  nb_cotes=0
  allocate(flag(1:nb_noeuds,1:nb_noeuds))      
          ! flag(n1,n2) contient en fin de boucle le numro de l'arte liant les sommets de numros n1 et n2
  flag=0
  do i=1,nb_trig
     n1=S(i,1); n2=S(i,2); n3=S(i,3)
     if (flag(n1,n2)==0.or.flag(n2,n1)==0) then ! alors arete non numerotee
        nb_cotes=nb_cotes+1
        flag(n1,n2)=nb_cotes; flag(n2,n1)=nb_cotes
     end if
     if (flag(n2,n3)==0.or.flag(n3,n2)==0) then ! alors arete non numerotee
        nb_cotes=nb_cotes+1
        flag(n2,n3)=nb_cotes; flag(n3,n2)=nb_cotes
     end if
     if (flag(n3,n1)==0.or.flag(n1,n3)==0) then ! alors arete non numerotee
        nb_cotes=nb_cotes+1
        flag(n3,n1)=nb_cotes; flag(n1,n3)=nb_cotes
     end if
  end do

  !2)-remplissage de e
  allocate(e(1:nb_cotes,2))    ! numros des noeuds extrmits de chaque arte
  do i=1,nb_trig
     n1=S(i,1); n2=S(i,2); n3=S(i,3)
     e(flag(n1,n2),1)=n1; e(flag(n1,n2),2)=n2
     e(flag(n2,n3),1)=n2; e(flag(n2,n3),2)=n3
     e(flag(n3,n1),1)=n3; e(flag(n3,n1),2)=n1
  end do

  !3)-remplissage de ar et trig
  allocate(ar(1:nb_trig,3))    ! numros des artes de chaque triangle
  allocate(trig(1:nb_cotes,2)) ! numro des triangles ayant l'arte i en commun
  trig=0
  do i=1,nb_trig
     n1=S(i,1); n2=S(i,2); n3=S(i,3)
     !-arete (n1,n2)
     ar(i,1)=flag(n1,n2)
     if (trig(ar(i,1),1)/=0) then
        trig(ar(i,1),2)=i
     else
        trig(ar(i,1),1)=i
     end if
     !-arete (n2,n3)
     ar(i,2)=flag(n2,n3)
     if (trig(ar(i,2),1)/=0) then
        trig(ar(i,2),2)=i
     else
        trig(ar(i,2),1)=i
     end if
     !-arete (n3,n1)
     ar(i,3)=flag(n3,n1)
     if (trig(ar(i,3),1)/=0) then
        trig(ar(i,3),2)=i
     else
        trig(ar(i,3),1)=i
     end if

  end do

  !--- calcul des aires des triangles et du "pas" du maillage
  allocate(aire(nb_trig),diam(nb_trig))
  do i=1,nb_trig
     A=P(s(i,1),:)
     B=P(s(i,2),:)
     C=P(s(i,3),:)
     aire(i)=0.5*abs((B(1)-A(1))*(C(2)-A(2))-(C(1)-A(1))*(B(2)-A(2)))
     AB=sqrt(sum((B-A)**2))
     BC=sqrt(sum((C-B)**2))
     AC=sqrt(sum((C-A)**2))
     diam(i)=2._pr*aire(i)/(AB+BC+AC)
  end do
  h=minval(diam)

  !--- calcul des centres des triangles
  allocate(xc(nb_trig),yc(nb_trig))
  do i=1,nb_trig
     xc(i)=1._pr/3._pr*(P(s(i,1),1)+P(s(i,2),1)+P(s(i,3),1))
     yc(i)=1._pr/3._pr*(P(s(i,1),2)+P(s(i,2),2)+P(s(i,3),2))
  end do

  !--- calcul des normales aux artes (definies comme pointant du triangle 1 vers le triangle 2)
  !-- on utilise le rangement des noeuds du triangle 1 dans le sens trigo pour orienter
  !-- l'arte dans le sens trigo,
  !-- puis faire une rotation de -pi/2 de l'arte
  allocate(normale(1:nb_cotes,1:2))
  do arete=1,nb_cotes

     iA=E(arete,1)
     iB=E(arete,2)

     tri1=trig(arete,1)

     if (s(tri1,1)/=iA.and.s(tri1,1)/=iB) then
        iA=s(tri1,2)
        iB=s(tri1,3)
     elseif (s(tri1,2)/=iA.and.s(tri1,2)/=iB) then
        iA=s(tri1,3)
        iB=s(tri1,1)
     else
        iA=s(tri1,1)
        iB=s(tri1,2)
     end if
     normale(arete,1)=P(iB,2)-P(iA,2)
     normale(arete,2)=-(P(iB,1)-P(iA,1))

  end do

  !--- calcul de la vitesse normale aux artes
  allocate(vnp(1:nb_cotes),vnm(1:nb_cotes))
  do arete=1,nb_cotes

     !-- indices des extremites
     ia=e(arete,1)
     ib=e(arete,2)

     !-- evaluation de la vitesse au milieu de l'arte
     v(1)=vx(0.5_pr*(P(iA,1)+P(iB,1)),0.5_pr*(P(iA,2)+P(iB,2)))
     v(2)=vy(0.5_pr*(P(iA,1)+P(iB,1)),0.5_pr*(P(iA,2)+P(iB,2)))

     !-- calcul des parties positives et ngatives de la vitesse normale
     vn=dot_product(v,normale(arete,:))
     vnp(arete)=0.5*(vn+abs(vn))
     vnm(arete)=0.5*(vn-abs(vn))

  end do

  !--- allocation des flux
  allocate(flux(1:nb_cotes))

  !--- donnees au bord
  fb0=0._pr ! utilise pour les exemples 1 et 2
!!$  fb1=1._pr ! utilise pour l'exemple 3

  !--- donnee initiale
  allocate(f(nb_trig),fnpun(nb_trig))
  f=ci(xc,yc)
  fnpun=0._pr

  !--- vitesse maximum et estimation du pas de temps
  vmax=0._pr
  do i=1,nb_noeuds
     vmax=max(vmax,sqrt((vx(P(i,1),P(i,2)))**2+(vy(P(i,1),P(i,2)))**2))
  end do
  cfl=0.9_pr

  !--- pas de temps
  invdt=0._pr
  do i=1,nb_trig
     !-- somme des vitesses normales positives
     svnp=0._pr
     do j=1,3
        svnp=svnp+vnp(ar(i,j))
     end do
     !-- invdt du triangle
     invdt=max(invdt,svnp/aire(i))
  end do
  dt=cfl/invdt

  !--- temps de simulation et parametres de sorties
  !-- exemple 1 : rotation rigide
  tmax=4._pr*acos(-1.0_pr) !- (pour faire deux tours)
  nmax=floor(tmax/dt) 
!!$  !-- exemples 2 et 3 : champ de vitesse a potentiel complexe autour d'un cylindre
!!$  tmax=20._pr/vmax
!!$  nmax=floor(tmax/dt) 

  print*,"nombre d'iterations en temps :",nmax
  pas_affichage=max(nmax/120,1)  !-- pour avoir 120 images, soit 5 secondes d'animation
  nfich=0

  !--- boucle en temps
  boucle_temps:do n=1,nmax

     !-- boucle sur les artes pour calculer les flux
     do arete=1,nb_cotes

        !-- indices des triangles voisins
        tri1=trig(arete,1)
        tri2=trig(arete,2)

        !-- valeur de f dans les triangles voisins
        fi=f(tri1)
        if (tri2==0) then ! alors arete est sur le bord
           if (refs(e(arete,1))==10.and.refs(e(arete,2))==10) then ! alors bord avec donnee rentrante
              fj=fb1
           else ! on ne fait rien rentrer
              fj=fb0
           end if
        else
           fj=f(tri2)
        end if

        !-- calcul du flux(f(tri1),f(tri2),ntri1tri2)
        flux(arete)=vnp(arete)*fi+vnm(arete)*fj

     end do

     !-- boucle sur les triangles
     boucle_triangle:do i=1,nb_trig

        !-- boucle sur les artes du triangle i
        R=0._pr

        boucle_arete:do k=1,3

           !-- numero global de l'arte
           arete=ar(i,k)

           !-- numeros globaux des deux triangles contenant arte
           tri1=trig(arete,1)
           tri2=trig(arete,2)

           !-- on cherche lequel est i
           if (tri1==i) then  ! (contribution + du flux)
              sf=1
           else               ! (contribution - du flux)
              sf=-1
           end if

           !-- mise a jour du residu
           R=R+sf*flux(arete)

        end do boucle_arete

        !-- mise a jour de fnpun
        fnpun(i)=f(i)-dt/aire(i)*R

     end do boucle_triangle

     !-- mise a jour de f
     f=fnpun

     !-- mise a jour du temps
     temps=temps+dt

     !-- sauvegarde resultats (pour visit)
     if (mod(n,pas_affichage)==0) then

        nfich=nfich+1
        write(numero,*) nfich
        if (nfich<=9) then
           numero='00'//trim(adjustl(numero))
        elseif (9<nfich.and.nfich<=99) then
           numero='0'//trim(adjustl(numero))
        end if

        call sortie_visit(numero,f,P,s)

        print*,n

     end if

  end do boucle_temps

  !----------------------------------------------------------------

end program conv2d_non_structure_cc
